from math import sqrt, fabs
from pyprocessing import *

# Global variables to hold constants
G            = -9.8
INF          = float('Inf')
windowWidth  = 1000
windowHeight = 200
rate         = 30
ballRadius   = 20
nBalls       = 30
tmax         = 40.0
hmax         = 40.0
nFrames      = 200
plateLo      = 1.0
plateHi      = 3.0
platePeriod  = 3.0
ballInitH    = 20.0
ballInitU    = -0.6

# Global variables to keep track of state
X            = 0
Y            = 0
clicked      = False
sawtooth     = False
xspeed       = 1
frame        = 0
history      = []

# Class for the vibrating plate
class Plate:

    """ A plate moving in sawtooth fashion with certain period between
    lo and hi.  For sawtooth motion, velocity of the plate is then
    v=2*(hi-lo)/period with alternating sign.  For non-sawtooth
    motion, a quadratic approximation to a sine function is used.  In
    both cases, the motion can be written in piecewise quadratic form:
    z=at^2+bt+c.
    """

    def __init__(self, lo, hi, T):
        """Initialize the vibrating plate's parameters: the lowest
        point is 'lo', the higher point is 'hi' and the period of the
        motion is 'T'."""
        self.z1 = min(lo, hi)
        self.z2 = max(lo, hi)
        self.dz = self.z2 - self.z1
        self.T  = float(fabs(T))
        self.V  = 2*self.dz/self.T
        self.A  = 8*self.dz/self.T**2
        self.sawtooth = sawtooth

    def piece(self, t):
        """What sawtooth piece are we on at time t?"""
        return int(2*t/self.T) if self.T != 0 else 0

    def turn(self, p):
        """Time of the plate's turning point in piece p for near-sinusoidal motion"""
        return self.T*(0.25+0.5*p)

    def a(self, p):
        if self.sawtooth:
            return 0.0
        else:
            if p%2 == 0:
                return -self.A
            else:
                return self.A

    def b(self, p):
        """Compute the velocity of the plate for piece p"""
        if self.sawtooth:
            if p%2 == 0:
                return self.V
            else: 
                return -self.V
        else:
            if p%2 == 0:
                return 2*self.A*self.turn(p)
            else: 
                return -2*self.A*self.turn(p)

    def c(self, p):
        """Compute the extrapolated position at time 0 for piece p"""
        if self.sawtooth:
            if p%2 == 0:
                return self.z1 - p*self.dz
            else: 
                return self.z1 + (p+1)*self.dz
        else:
            if p%2 == 0:
                return self.z2 - self.A*self.turn(p)**2
            else: 
                return self.z1 + self.A*self.turn(p)**2

    def z(self, t):
        """compute the position of the plate at time t"""
        p = self.piece(t)
        return self.a(p)*t**2 + self.b(p)*t + self.c(p)

    def v(self, t):
        """Compute the velocity of the plate at time t"""
        p = self.piece(t)
        return 2*self.a(p)*t + self.b(p)

# Class for the ball object

class Ball:

    """A ball moving under gravity. The motion of the ball (between
    bounces) can be written in piecewise quadratic form: 
    h=at^2+bt+c."""

    def __init__(self, t0, h0, u0):
        self.t0 = t0
        self.h0 = h0
        self.u0 = u0

    def update(self, t0, h0, u0):
        """Update values in the ball"""
        self.t0 = t0
        self.h0 = h0
        self.u0 = u0

    def a(self):
        return 0.5*G

    def b(self):
        return self.u0 - G * self.t0

    def c(self):
        return self.h0 - self.u0*self.t0 + 0.5*G*self.t0**2

    def h(self, t):
        """compute the height of the plate at timet """
        return self.a()*t**2 + self.b()*t + self.c()

    def u(self, t):
        """Compute the velocity of the plate at time t"""
        return 2*self.a()*t + self.b()
   
    def whenat(self, htarget):
        """compute the last time when the ball's height value equals htarget"""
        a = self.a()
        b = self.b()
        c = self.c() - htarget
        d = b**2 - 4*a*c
        if d == 0:
            return [- b/(2*a)]
        elif d > 0:
            t1 = (-b + sqrt(d))/(2*a)
            t2 = (-b - sqrt(d))/(2*a)
            return [min(t1,t2), max(t1,t2)]
        else:
            return [INF]

def futureBounceTimeForPiece(ball, plate, p):
    """Instant after t at which the falling 'ball' will hit the moving 'plate' for piece p"""
    a  = ball.a() - plate.a(p)
    b  = ball.b() - plate.b(p)
    c  = ball.c() - plate.c(p)
    d  = b**2 - 4*a*c
    if d == 0:
        t = - b/(2*a)        
        if plate.piece(t) != p: t = INF
    elif d > 0:
        t1 = (-b + sqrt(d))/(2*a)
        t2 = (-b - sqrt(d))/(2*a)
        t = max(t1,t2)
        if plate.piece(t) != p: t = INF
    else:
        t = INF
    return t  

def computeBounceTime(ball, plate, t):
    """Instant after 't' at which the falling 'ball' will hit the moving 'plate'"""
    # bracket the time between t1 and t2
    t1 = t # lower bracket is current time (too broad in general, but
           # correct if t is a bounce, which is the typical case here)
    #first determine the time when the ball would hit z=lo or hi
    eps = 1e-6  # tolerance on bounce being after 't'
    t2 = -INF
    for tb in ball.whenat(plate.z1):
        if tb != INF and tb >= t-eps and tb > t2: 
            t2 = tb
    for tb in ball.whenat(plate.z2):
        if tb != INF and tb >= t-eps and tb > t2: 
            t2 = tb
    if t2==-INF:
        tb = INF
    else:
        # determine what pieces get traversed in the bracketed interval
        s1 = plate.piece(t1)        
        s2 = plate.piece(t2)
        # find the first time at which we may bounce upon that a piece in the future.
        tb = INF
        for s in xrange(s1,s2+1):
            tbs = futureBounceTimeForPiece(ball, plate, s)
            if tbs >= t-eps and tbs < tb: 
                tb = tbs
    return tb

# A class to combing one Ball with one Vibrating Plate
class BouncingBallSystem:

    """Combination of a moving plate and a falling ball"""

    def setmeup(self, lo, hi, T, tmax, h0, u0):
        """Sets parameters into an existing BouncingBallSystem"""
        self.plate = Plate(lo=lo, hi=hi, T=T)
        self.ball = Ball(t0=0.0, h0=h0, u0=u0)
        self.t = 0.0
        self.tmax = tmax
        self.tn = INF
        self.compute_tn()

    def __init__(self, lo, hi, T, tmax, h0, u0):
        """Initialize BouncingBallSystem"""
        self.setmeup(lo, hi, T, tmax, h0, u0)
        # Save these parameters
        self.save = {'lo':lo, 'hi':hi, 'T':T, 'tmax':tmax, 'h0':h0, 'u0':u0}

    def restore(self):
        """Restore the original state."""
        self.setmeup(**self.save)
        
    def compute_tn(self):
        self.tn = computeBounceTime(self.ball, self.plate, self.t)

    def performNextCollision(self):
        self.ball.update(self.tn, self.ball.h(self.tn),
                         2*self.plate.v(self.tn) - self.ball.u(self.tn))
        self.t = self.tn
        self.compute_tn()

    def getLastCollision(self):
        return self.t

    def getBallHeight(self, t):
        while (t > self.tn):
            self.performNextCollision()
        return self.ball.h(t)
        
    def getPlateHeight(self, t):
        while (t > self.tn):
            self.performNextCollision()
        return self.plate.z(t)

    def getTmax(self):
        return self.tmax

    
def fromto(i, n, x1, x2):
    return x1 + ((x2 - x1)*i)/(n-1)

def setup():
    """Function that gets called by 'pyprocessing' when the application starts"""
    global system
    size(windowWidth, windowHeight)
    frameRate(rate)
    system = [ BouncingBallSystem(lo=plateLo, hi=plateHi, T=platePeriod, 
                                   tmax=tmax, h0=fromto(i,nBalls,ballInitH*0.95,ballInitH*1.05), u0=ballInitU)
               for i in range(nBalls)]

def mouseMoved():
    """Function that gets called by 'pyprocessing' when the mouse moves"""
    # Records the mouse pointer position if mouse not yet clicked
    global clicked, X, Y
    if not clicked:
        X = mouse.x
        Y = mouse.y

def mouseClicked():
    """Function that gets called by 'pyprocessing' when the mouse button is clicked"""
    # Records the mouse pointer position and forces a restart if clicked again
    global frame, clicked, X, Y
    X = mouse.x
    Y = mouse.y
    if clicked:
        frame=100000 # forces a restart in draw
    else:
        clicked = True

def keyPressed():
    """Called when a key is pressed; SPACE toggles the motion of the plate"""
    global sawtooth
    if key.char == " ":
        sawtooth = not sawtooth

def draw():
    global xspeed, frame, system, history, clicked, X, Y
    midindex = len(system)/2
    if frame < nFrames:
        # time of this frame
        t  = system[midindex].getTmax()*frame/((float)(nFrames-1))
        # height of all the balls in the system at this instant of time
        h  = [s.getBallHeight(t) for s in system]
        # height of the floor at this instant of time
        f  = system[midindex].getPlateHeight(t)
        # window x and y coordinates of the balls
        px = ballRadius + (windowWidth-2*ballRadius)*t*xspeed/system[midindex].getTmax()
        py = [-ballRadius/2 + windowHeight*(hmax-hp)/hmax for hp in h]
        # add these to the trajectory list
        history.append([px,py])
        # window x and y coordinate of the floor
        fx = windowWidth
        fy = windowHeight*(hmax-f)/hmax
        # diameter of the ball, which may be squished to an ellipse
        dx = 2*ballRadius
        dy = 2*ballRadius
        if (py[0] + dy/2 > fy):
            dy = 2*(fy-py[0])
        dx = sqrt(8*ballRadius**3/dy)
        # draw the background
        background(100,140,200)
        # print the time and controls
        fill(0)
        text("t="+str(int(t)),20,20)
        text("SPACE toggles motion",20,40)
        text("ESC quits",20,60)
        # print the next plate motion
        if sawtooth:
            text("next plate motion: sawtooth", windowWidth/2, 20)
        else:
            text("next plate motion: near-sinusoidal", windowWidth/2, 20)
        # draw the trajectories
        stroke(100,255,100)
        strokeWeight(1)
        noFill()
        histsize=len(history[0][1])
        for j in xrange(histsize):
            beginShape()
            for i in xrange(len(history)):
                vertex(history[i][0],history[i][1][j])
            endShape()
        # draw the ball
        stroke(0)
        fill(0)
        ellipse(px, py[0], dx, dy)
        # draw the plate
        stroke(255)
        strokeWeight(4)
        line(0, fy, fx, fy)
        # interactive arrow
        if clicked:
            stroke(255,0,0)
            strokeWeight(3)
        else:
            stroke(0)
            strokeWeight(3)
        line(0, Y, X, Y)
        line(X, Y, X-10, Y-10)
        line(X, Y, X-10, Y+10)
    #
    if frame < nFrames - 1 and t*xspeed < system[midindex].getTmax(): 
        # next frame if there are frames left
        frame = frame + 1
    else:
        if clicked:
            # based on mouse position when clicked, select a new height
            newh0 = hmax*(1.0 - Y/float(windowHeight))
            if newh0 < 3: 
                newh0 = 3+0.01
            system=[BouncingBallSystem(lo=plateLo,
                                       hi=plateHi,
                                       T=platePeriod, 
                                       tmax=tmax, 
                                       h0=fromto(i,nBalls,newh0*0.95,newh0*1.05),
                                       u0=ballInitU)
                    for i in range(nBalls)]
            xspeed = 1+int(10*X/float(windowWidth));
            clicked = False
        else:
            for s in system:
                s.restore()
        frame = 0
        history = []

# Start main processing loop
run()    

